from unittest.mock import patch
from django.test import TestCase

from esi.exceptions import HTTPNotModified

from app_utils.testdata_factories import UserMainFactory

from ..models import AllianceToken, AllianceContactLabel, AllianceContact, CorporationContact, CorporationContactLabel, CorporationToken
from ..tasks import update_alliance_contacts, update_corporation_contacts

from .utils import SimpleAttributeDict


class TestUpdateAllianceContacts(TestCase):

    @classmethod
    def setUpTestData(cls):
        user = UserMainFactory()
        cls.alliance = user.profile.main_character.alliance
        token = user.token_set.first()

        cls.token = AllianceToken.objects.create(
            alliance=cls.alliance,
            token=token
        )

        cls.label_data = [
            SimpleAttributeDict({"label_id": 1, "label_name": "Test Label 1"}),
            SimpleAttributeDict({"label_id": 2, "label_name": "Test Label 2"}),
        ]

        cls.contact_data = [
            SimpleAttributeDict({
                "contact_id": 2,
                "contact_type": "corporation",
                'label_ids': [1, 2],
                'standing': 5.0
            }),
        ]

    @patch("aa_contacts.models.ContactTokenQueryset.with_valid_tokens")
    def test_token_doesnt_exist(self, mock_with_valid_tokens):
        mock_with_valid_tokens.return_value = AllianceToken.objects.all()

        self.token.delete()
        with self.assertRaises(ValueError):
            update_alliance_contacts(self.alliance.alliance_id)

    @patch("aa_contacts.tasks.AllianceContactUpdater.get_labels_data")
    @patch("aa_contacts.tasks.AllianceContactUpdater.get_contacts_data")
    @patch("aa_contacts.tasks.group.delay")
    @patch("aa_contacts.models.ContactTokenQueryset.with_valid_tokens")
    def test_ok(self, mock_with_valid_tokens, mock_delay, mock_get_contacts_data, mock_get_labels_data):
        mock_with_valid_tokens.return_value = AllianceToken.objects.all()
        mock_delay.return_value = None

        mock_get_labels_data.return_value = self.label_data
        mock_get_contacts_data.return_value = self.contact_data

        update_alliance_contacts(self.alliance.alliance_id)

        self.assertEqual(AllianceContact.objects.count(), 1)
        self.assertEqual(AllianceContactLabel.objects.count(), 2)

    @patch("aa_contacts.tasks.AllianceContactUpdater.get_labels_data")
    @patch("aa_contacts.tasks.AllianceContactUpdater.get_contacts_data")
    @patch("aa_contacts.tasks.group.delay")
    @patch("aa_contacts.models.ContactTokenQueryset.with_valid_tokens")
    def test_update(self, mock_with_valid_tokens, mock_delay, mock_get_contacts_data, mock_get_labels_data):
        mock_with_valid_tokens.return_value = AllianceToken.objects.all()
        mock_delay.return_value = None

        mock_get_labels_data.return_value = self.label_data
        mock_get_contacts_data.return_value = self.contact_data

        update_alliance_contacts(self.alliance.alliance_id)

        self.assertEqual(AllianceContact.objects.count(), 1)
        self.assertEqual(AllianceContactLabel.objects.count(), 2)

        self.label_data.pop()
        self.contact_data[0]['label_ids'].pop()

        update_alliance_contacts(self.alliance.alliance_id)

        self.assertEqual(AllianceContact.objects.count(), 1)
        self.assertEqual(AllianceContactLabel.objects.count(), 1)

        mock_get_contacts_data.return_value = []

        contact = AllianceContact.objects.first()
        contact.notes = "Test"
        contact.save()

        update_alliance_contacts(self.alliance.alliance_id)

        self.assertEqual(AllianceContact.objects.count(), 1)
        contact.refresh_from_db()
        self.assertEqual(contact.standing, 0.0)
        self.assertEqual(contact.labels.count(), 0)

        contact.notes = ""
        contact.save()

        update_alliance_contacts(self.alliance.alliance_id)

        self.assertEqual(AllianceContact.objects.count(), 0)

    @patch("aa_contacts.tasks.AllianceContactUpdater.get_labels_data")
    @patch("aa_contacts.tasks.AllianceContactUpdater.get_contacts_data")
    @patch("aa_contacts.tasks.group.delay")
    @patch("aa_contacts.models.ContactTokenQueryset.with_valid_tokens")
    def test_labels_not_modified(self, mock_with_valid_tokens, mock_delay, mock_get_contacts_data, mock_get_labels_data):
        mock_with_valid_tokens.return_value = AllianceToken.objects.all()
        mock_delay.return_value = None

        mock_get_labels_data.side_effect = HTTPNotModified(status_code=304, headers={})
        mock_get_contacts_data.return_value = self.contact_data

        AllianceContactLabel.objects.bulk_create([
            AllianceContactLabel(
                alliance=self.alliance,
                label_id=1,
                label_name="I am Groot 1",
            ),
            AllianceContactLabel(
                alliance=self.alliance,
                label_id=2,
                label_name="I am Groot 2",
            ),
            AllianceContactLabel(
                alliance=self.alliance,
                label_id=3,
                label_name="I am Groot 3",
            ),
        ])

        update_alliance_contacts(self.alliance.alliance_id)

        self.assertEqual(AllianceContact.objects.count(), 1)
        self.assertEqual(AllianceContactLabel.objects.count(), 3)
        self.assertEqual(
            AllianceContactLabel.objects.get(alliance=self.alliance, label_id=1).label_name,
            "I am Groot 1"
        )
        self.assertEqual(
            AllianceContactLabel.objects.get(alliance=self.alliance, label_id=2).label_name,
            "I am Groot 2"
        )
        self.assertEqual(
            AllianceContactLabel.objects.get(alliance=self.alliance, label_id=3).label_name,
            "I am Groot 3"
        )

    @patch("aa_contacts.tasks.AllianceContactUpdater.get_labels_data")
    @patch("aa_contacts.tasks.AllianceContactUpdater.get_contacts_data")
    @patch("aa_contacts.tasks.group.delay")
    @patch("aa_contacts.models.ContactTokenQueryset.with_valid_tokens")
    def test_contacts_not_modified(self, mock_with_valid_tokens, mock_delay, mock_get_contacts_data, mock_get_labels_data):
        mock_with_valid_tokens.return_value = AllianceToken.objects.all()
        mock_delay.return_value = None

        mock_get_labels_data.return_value = self.label_data
        mock_get_contacts_data.side_effect = HTTPNotModified(status_code=304, headers={})

        AllianceContact.objects.create(
            alliance=self.alliance,
            contact_id=3,
            contact_type="corporation",
            standing=3.0,
        )

        update_alliance_contacts(self.alliance.alliance_id)

        self.assertEqual(AllianceContact.objects.count(), 1)
        self.assertEqual(AllianceContactLabel.objects.count(), 2)

        contact = AllianceContact.objects.first()
        self.assertEqual(contact.contact_id, 3)

    @patch("aa_contacts.tasks.AllianceContactUpdater.get_labels_data")
    @patch("aa_contacts.tasks.AllianceContactUpdater.get_contacts_data")
    @patch("aa_contacts.tasks.group.delay")
    @patch("aa_contacts.models.ContactTokenQueryset.with_valid_tokens")
    def test_both_not_modified(self, mock_with_valid_tokens, mock_delay, mock_get_contacts_data, mock_get_labels_data):
        mock_with_valid_tokens.return_value = AllianceToken.objects.all()
        mock_delay.return_value = None

        mock_get_labels_data.side_effect = HTTPNotModified(status_code=304, headers={})
        mock_get_contacts_data.side_effect = HTTPNotModified(status_code=304, headers={})

        label1 = AllianceContactLabel(
            alliance=self.alliance,
            label_id=1,
            label_name="I am Groot 1",
        )
        label1.save()

        label2 = AllianceContactLabel(
            alliance=self.alliance,
            label_id=2,
            label_name="I am Groot 2",
        )
        label2.save()

        label3 = AllianceContactLabel(
            alliance=self.alliance,
            label_id=3,
            label_name="I am Groot 3",
        )
        label3.save()

        contact: AllianceContact = AllianceContact(
            alliance=self.alliance,
            contact_id=3,
            contact_type="corporation",
            standing=3.0,
        )
        contact.save()
        contact.labels.add(label1, label2, label3)

        update_alliance_contacts(self.alliance.alliance_id)
        self.assertEqual(AllianceContact.objects.count(), 1)
        self.assertEqual(AllianceContactLabel.objects.count(), 3)
        self.assertEqual(
            AllianceContactLabel.objects.get(alliance=self.alliance, label_id=1).label_name,
            "I am Groot 1"
        )
        self.assertEqual(
            AllianceContactLabel.objects.get(alliance=self.alliance, label_id=2).label_name,
            "I am Groot 2"
        )
        self.assertEqual(
            AllianceContactLabel.objects.get(alliance=self.alliance, label_id=3).label_name,
            "I am Groot 3"
        )

        contact.refresh_from_db()
        self.assertEqual(contact.contact_id, 3)


class TestUpdateCorporationContacts(TestCase):

    @classmethod
    def setUpTestData(cls):
        user = UserMainFactory()
        cls.corporation = user.profile.main_character.corporation
        token = user.token_set.first()

        cls.token = CorporationToken.objects.create(
            corporation=cls.corporation,
            token=token
        )

        cls.label_data = [
            SimpleAttributeDict({"label_id": 1, "label_name": "Test Label 1"}),
            SimpleAttributeDict({"label_id": 2, "label_name": "Test Label 2"}),
        ]

        cls.contact_data = [
            SimpleAttributeDict({
                "contact_id": 2,
                "contact_type": "alliance",
                'label_ids': [1, 2],
                'standing': 5.0
            }),
        ]

    @patch("aa_contacts.models.ContactTokenQueryset.with_valid_tokens")
    def test_token_doesnt_exist(self, mock_with_valid_tokens):
        mock_with_valid_tokens.return_value = CorporationToken.objects.all()

        self.token.delete()
        with self.assertRaises(ValueError):
            update_corporation_contacts(self.corporation.corporation_id)

    @patch("aa_contacts.tasks.CorporationContactUpdater.get_labels_data")
    @patch("aa_contacts.tasks.CorporationContactUpdater.get_contacts_data")
    @patch("aa_contacts.tasks.group.delay")
    @patch("aa_contacts.models.ContactTokenQueryset.with_valid_tokens")
    def test_ok(self, mock_with_valid_tokens, mock_delay, mock_get_contacts_data, mock_get_labels_data):
        mock_with_valid_tokens.return_value = CorporationToken.objects.all()
        mock_delay.return_value = None

        mock_get_labels_data.return_value = self.label_data
        mock_get_contacts_data.return_value = self.contact_data

        update_corporation_contacts(self.corporation.corporation_id)

        self.assertEqual(CorporationContact.objects.count(), 1)
        self.assertEqual(CorporationContactLabel.objects.count(), 2)

    @patch("aa_contacts.tasks.CorporationContactUpdater.get_labels_data")
    @patch("aa_contacts.tasks.CorporationContactUpdater.get_contacts_data")
    @patch("aa_contacts.tasks.group.delay")
    @patch("aa_contacts.models.ContactTokenQueryset.with_valid_tokens")
    def test_update(self, mock_with_valid_tokens, mock_delay, mock_get_contacts_data, mock_get_labels_data):
        mock_with_valid_tokens.return_value = CorporationToken.objects.all()
        mock_delay.return_value = None

        mock_get_labels_data.return_value = self.label_data
        mock_get_contacts_data.return_value = self.contact_data

        update_corporation_contacts(self.corporation.corporation_id)

        self.assertEqual(CorporationContact.objects.count(), 1)
        self.assertEqual(CorporationContactLabel.objects.count(), 2)

        self.label_data.pop()
        self.contact_data[0]['label_ids'].pop()

        update_corporation_contacts(self.corporation.corporation_id)

        self.assertEqual(CorporationContact.objects.count(), 1)
        self.assertEqual(CorporationContactLabel.objects.count(), 1)

        mock_get_contacts_data.return_value = []

        contact = CorporationContact.objects.first()
        contact.notes = "Test"
        contact.save()

        update_corporation_contacts(self.corporation.corporation_id)

        self.assertEqual(CorporationContact.objects.count(), 1)
        contact.refresh_from_db()
        self.assertEqual(contact.standing, 0.0)
        self.assertEqual(contact.labels.count(), 0)

        contact.notes = ""
        contact.save()

        update_corporation_contacts(self.corporation.corporation_id)

        self.assertEqual(CorporationContact.objects.count(), 0)

    @patch("aa_contacts.tasks.CorporationContactUpdater.get_labels_data")
    @patch("aa_contacts.tasks.CorporationContactUpdater.get_contacts_data")
    @patch("aa_contacts.tasks.group.delay")
    @patch("aa_contacts.models.ContactTokenQueryset.with_valid_tokens")
    def test_labels_not_modified(self, mock_with_valid_tokens, mock_delay, mock_get_contacts_data, mock_get_labels_data):
        mock_with_valid_tokens.return_value = CorporationToken.objects.all()
        mock_delay.return_value = None

        mock_get_labels_data.side_effect = HTTPNotModified(status_code=304, headers={})
        mock_get_contacts_data.return_value = self.contact_data

        CorporationContactLabel.objects.bulk_create([
            CorporationContactLabel(
                corporation=self.corporation,
                label_id=1,
                label_name="I am Groot 1",
            ),
            CorporationContactLabel(
                corporation=self.corporation,
                label_id=2,
                label_name="I am Groot 2",
            ),
            CorporationContactLabel(
                corporation=self.corporation,
                label_id=3,
                label_name="I am Groot 3",
            ),
        ])

        update_corporation_contacts(self.corporation.corporation_id)

        self.assertEqual(CorporationContact.objects.count(), 1)
        self.assertEqual(CorporationContactLabel.objects.count(), 3)
        self.assertEqual(
            CorporationContactLabel.objects.get(corporation=self.corporation, label_id=1).label_name,
            "I am Groot 1"
        )
        self.assertEqual(
            CorporationContactLabel.objects.get(corporation=self.corporation, label_id=2).label_name,
            "I am Groot 2"
        )
        self.assertEqual(
            CorporationContactLabel.objects.get(corporation=self.corporation, label_id=3).label_name,
            "I am Groot 3"
        )

    @patch("aa_contacts.tasks.CorporationContactUpdater.get_labels_data")
    @patch("aa_contacts.tasks.CorporationContactUpdater.get_contacts_data")
    @patch("aa_contacts.tasks.group.delay")
    @patch("aa_contacts.models.ContactTokenQueryset.with_valid_tokens")
    def test_contacts_not_modified(self, mock_with_valid_tokens, mock_delay, mock_get_contacts_data, mock_get_labels_data):
        mock_with_valid_tokens.return_value = CorporationToken.objects.all()
        mock_delay.return_value = None

        mock_get_labels_data.return_value = self.label_data
        mock_get_contacts_data.side_effect = HTTPNotModified(status_code=304, headers={})

        CorporationContact.objects.create(
            corporation=self.corporation,
            contact_id=3,
            contact_type="alliance",
            standing=3.0,
        )

        update_corporation_contacts(self.corporation.corporation_id)

        self.assertEqual(CorporationContact.objects.count(), 1)
        self.assertEqual(CorporationContactLabel.objects.count(), 2)

        contact = CorporationContact.objects.first()
        self.assertEqual(contact.contact_id, 3)

    @patch("aa_contacts.tasks.CorporationContactUpdater.get_labels_data")
    @patch("aa_contacts.tasks.CorporationContactUpdater.get_contacts_data")
    @patch("aa_contacts.tasks.group.delay")
    @patch("aa_contacts.models.ContactTokenQueryset.with_valid_tokens")
    def test_both_not_modified(self, mock_with_valid_tokens, mock_delay, mock_get_contacts_data, mock_get_labels_data):
        mock_with_valid_tokens.return_value = CorporationToken.objects.all()
        mock_delay.return_value = None

        mock_get_labels_data.side_effect = HTTPNotModified(status_code=304, headers={})
        mock_get_contacts_data.side_effect = HTTPNotModified(status_code=304, headers={})

        label1 = CorporationContactLabel(
            corporation=self.corporation,
            label_id=1,
            label_name="I am Groot 1",
        )
        label1.save()

        label2 = CorporationContactLabel(
            corporation=self.corporation,
            label_id=2,
            label_name="I am Groot 2",
        )
        label2.save()

        label3 = CorporationContactLabel(
            corporation=self.corporation,
            label_id=3,
            label_name="I am Groot 3",
        )
        label3.save()

        contact: CorporationContact = CorporationContact(
            corporation=self.corporation,
            contact_id=3,
            contact_type="alliance",
            standing=3.0,
        )
        contact.save()
        contact.labels.add(label1, label2, label3)

        update_corporation_contacts(self.corporation.corporation_id)
        self.assertEqual(CorporationContact.objects.count(), 1)
        self.assertEqual(CorporationContactLabel.objects.count(), 3)
        self.assertEqual(
            CorporationContactLabel.objects.get(corporation=self.corporation, label_id=1).label_name,
            "I am Groot 1"
        )
        self.assertEqual(
            CorporationContactLabel.objects.get(corporation=self.corporation, label_id=2).label_name,
            "I am Groot 2"
        )
        self.assertEqual(
            CorporationContactLabel.objects.get(corporation=self.corporation, label_id=3).label_name,
            "I am Groot 3"
        )

        contact.refresh_from_db()
        self.assertEqual(contact.contact_id, 3)



from esi.openapi_clients import ESIClientProvider

from . import (
    __version__ as app_version,
    __app_name_ua__ as app_name_ua,
    __github_url__ as github_url,
    __esi_compatibility_date__ as esi_compatibility_date
)

esi = ESIClientProvider(
    compatibility_date=esi_compatibility_date,
    ua_appname=app_name_ua,
    ua_version=app_version,
    ua_url=github_url,
    tags=["Contacts"]
)



from random import randint

from celery import shared_task, group
from celery_once import QueueOnce

from django.db import transaction
from django.utils import timezone

from allianceauth.eveonline.models import EveAllianceInfo, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger

from esi.models import Token
from esi.exceptions import HTTPNotModified

from .app_settings import TASK_JITTER
from .models import AllianceContact, AllianceContactLabel, AllianceToken, CorporationToken, CorporationContact, CorporationContactLabel
from .providers import esi

logger = get_extension_logger(__name__)


@shared_task(base=QueueOnce, once={'graceful': True})
def load_alliance_contact_name(contact_pk: int):
    contact = AllianceContact.objects.get(pk=contact_pk)
    contact.contact_name


@shared_task(base=QueueOnce, once={'graceful': True})
def load_corporation_contact_name(contact_pk: int):
    contact = CorporationContact.objects.get(pk=contact_pk)
    contact.contact_name


class BaseContactUpdater:

    @classmethod
    def get_entity_object(cls, entity_id: int) -> EveCorporationInfo | EveAllianceInfo:
        raise NotImplementedError

    @classmethod
    def get_entity_token(cls, entity: EveCorporationInfo | EveAllianceInfo) -> CorporationToken | AllianceToken:
        raise NotImplementedError

    @classmethod
    def get_labels_data(cls, entity_id: int, token: Token):
        raise NotImplementedError

    @classmethod
    def get_contacts_data(cls, entity_id: int, token: Token):
        raise NotImplementedError

    @classmethod
    def get_contacts_model(cls):
        raise NotImplementedError

    @classmethod
    def get_entity_selector(cls, entity: EveCorporationInfo | EveAllianceInfo) -> dict[str, EveCorporationInfo | EveAllianceInfo]:
        raise NotImplementedError

    @classmethod
    def get_labels_model(cls):
        raise NotImplementedError

    @classmethod
    def get_labels_related_selector(cls, pks: list[int]) -> dict[str, list[int]]:
        raise NotImplementedError

    @classmethod
    def get_load_contact_name_task(cls):
        raise NotImplementedError

    @classmethod
    def update_contacts(cls, entity_id: int):
        contacts_to_load = []

        entity = cls.get_entity_object(entity_id)
        entity_token = cls.get_entity_token(entity)

        try:
            labels_data = cls.get_labels_data(entity_id, entity_token.token)
        except HTTPNotModified:
            update_labels = False
        else:
            labels = {label.label_id: label.label_name for label in labels_data}
            update_labels = True

        try:
            contacts_data = cls.get_contacts_data(entity_id, entity_token.token)
        except HTTPNotModified:
            update_contacts = False
        else:
            contact_ids = {
                contact.contact_id: {
                    'contact_type': contact.contact_type,
                    'label_ids': contact.label_ids or [],
                    'standing': contact.standing
                } for contact in contacts_data
            }
            update_contacts = True

        with transaction.atomic():
            label_objects = {}

            if update_labels:
                cls.get_labels_model().objects.filter(
                    **cls.get_entity_selector(entity)
                ).exclude(
                    label_id__in=labels.keys()
                ).delete()

                for label_id, label_name in labels.items():
                    label, _ = cls.get_labels_model().objects.update_or_create(
                        **cls.get_entity_selector(entity),
                        label_id=label_id,
                        defaults={'label_name': label_name}
                    )

                    label_objects[label_id] = label
            elif update_contacts:
                existing_labels = cls.get_labels_model().objects.filter(
                    **cls.get_entity_selector(entity)
                )
                label_objects = {label.label_id: label for label in existing_labels}

            if update_contacts:
                missing_contacts = cls.get_contacts_model().objects.filter(
                    **cls.get_entity_selector(entity)
                ).exclude(
                    contact_id__in=contact_ids.keys()
                )

                missing_contacts.filter(notes='').delete()
                cls.get_contacts_model().labels.through.objects.filter(
                    **cls.get_labels_related_selector(missing_contacts.values('pk'))
                ).delete()
                missing_contacts.update(standing=0.0)

                for contact_id, contact_data in contact_ids.items():
                    contact, _ = cls.get_contacts_model().objects.update_or_create(
                        **cls.get_entity_selector(entity),
                        contact_id=contact_id,
                        defaults={
                            'contact_type': contact_data['contact_type'],
                            'standing': contact_data['standing']
                        }
                    )

                    contact.labels.clear()
                    if contact_data['label_ids'] is not None:
                        contact.labels.set([label_objects[label_id] for label_id in contact_data['label_ids']])

                    contacts_to_load.append(contact.pk)

            entity_token.last_update = timezone.now()
            entity_token.save()

        contacts_to_load = cls.get_contacts_model().filter_missing_contact_name(contacts_to_load)
        if len(contacts_to_load) > 0:
            group(cls.get_load_contact_name_task().si(pk) for pk in contacts_to_load).delay()


class CorporationContactUpdater(BaseContactUpdater):

    @classmethod
    def get_entity_object(cls, entity_id: int) -> EveCorporationInfo:
        try:
            corporation = EveCorporationInfo.objects.get(corporation_id=entity_id)
        except EveCorporationInfo.DoesNotExist:
            corporation = EveCorporationInfo.objects.create_corporation(entity_id)

        return corporation

    @classmethod
    def get_entity_token(cls, entity: EveCorporationInfo) -> CorporationToken:
        try:
            return CorporationToken.objects.with_valid_tokens().select_related('token').get(corporation=entity)
        except CorporationToken.DoesNotExist:
            raise ValueError(f"No valid token found for corporation {entity}")

    @classmethod
    def get_labels_data(cls, entity_id: int, token: Token):
        return (
            esi.client
            .Contacts
            .GetCorporationsCorporationIdContactsLabels(
                corporation_id=entity_id,
                token=token
            )
            .results()
        )

    @classmethod
    def get_contacts_data(cls, entity_id: int, token: Token):
        return (
            esi.client
            .Contacts
            .GetCorporationsCorporationIdContacts(
                corporation_id=entity_id,
                token=token
            )
            .results()
        )

    @classmethod
    def get_contacts_model(cls):
        return CorporationContact

    @classmethod
    def get_entity_selector(cls, entity: EveCorporationInfo) -> dict[str, EveCorporationInfo]:
        return {"corporation": entity}

    @classmethod
    def get_labels_model(cls):
        return CorporationContactLabel

    @classmethod
    def get_labels_related_selector(cls, pks: list[int]) -> dict[str, list[int]]:
        return {"corporationcontact_id__in": pks}

    @classmethod
    def get_load_contact_name_task(cls):
        return load_corporation_contact_name


class AllianceContactUpdater(BaseContactUpdater):

    @classmethod
    def get_entity_object(cls, entity_id: int) -> EveAllianceInfo:
        try:
            alliance = EveAllianceInfo.objects.get(alliance_id=entity_id)
        except EveAllianceInfo.DoesNotExist:
            alliance = EveAllianceInfo.objects.create_alliance(entity_id)

        return alliance

    @classmethod
    def get_entity_token(cls, entity: EveAllianceInfo) -> AllianceToken:
        try:
            return AllianceToken.objects.with_valid_tokens().select_related('token').get(alliance=entity)
        except AllianceToken.DoesNotExist:
            raise ValueError(f"No valid token found for alliance {entity}")

    @classmethod
    def get_labels_data(cls, entity_id: int, token: Token):
        return (
            esi.client
            .Contacts
            .GetAlliancesAllianceIdContactsLabels(
                alliance_id=entity_id,
                token=token
            ).results()
        )

    @classmethod
    def get_contacts_data(cls, entity_id: int, token: Token):
        return (
            esi.client
            .Contacts
            .GetAlliancesAllianceIdContacts(
                alliance_id=entity_id,
                token=token
            ).results()
        )

    @classmethod
    def get_contacts_model(cls):
        return AllianceContact

    @classmethod
    def get_entity_selector(cls, entity: EveAllianceInfo) -> dict[str, EveAllianceInfo]:
        return {"alliance": entity}

    @classmethod
    def get_labels_model(cls):
        return AllianceContactLabel

    @classmethod
    def get_labels_related_selector(cls, pks: list[int]) -> dict[str, list[int]]:
        return {"alliancecontact_id__in": pks}

    @classmethod
    def get_load_contact_name_task(cls):
        return load_alliance_contact_name


@shared_task(base=QueueOnce, once={'graceful': True})
def update_alliance_contacts(alliance_id: int):
    AllianceContactUpdater.update_contacts(alliance_id)


@shared_task(base=QueueOnce, once={'graceful': True})
def update_corporation_contacts(corporation_id: int):
    CorporationContactUpdater.update_contacts(corporation_id)


@shared_task
def update_all_alliances_contacts():
    group(
        update_alliance_contacts.si(alliance_token.alliance.alliance_id).set(countdown=randint(0, TASK_JITTER))
        for alliance_token in AllianceToken.objects.with_valid_tokens().select_related('alliance')
    ).delay()


@shared_task
def update_all_corporations_contacts():
    group(
        update_corporation_contacts.si(corporation_token.corporation.corporation_id).set(countdown=randint(0, TASK_JITTER))
        for corporation_token in CorporationToken.objects.with_valid_tokens().select_related('corporation')
    ).delay()


@shared_task
def update_all_contacts():
    update_all_alliances_contacts.delay()
    update_all_corporations_contacts.delay()



import copy


class SimpleAttributeDict(dict):
    def __getattr__(self, item):
        return self[item]

    def __setattr__(self, key, value):
        self[key] = value

    def __deepcopy__(self, memo):
        copied = SimpleAttributeDict()
        for k, v in self.items():
            copied[copy.deepcopy(k, memo)] = copy.deepcopy(v, memo)
        return copied



from ninja import Router

from django.contrib.auth.models import User

from allianceauth.authentication.models import CharacterOwnership
from aa_contacts.models import CorporationToken

from ..schema import CorporationTokenSchema


router = Router()


@router.get("/", response=list[CorporationTokenSchema])
def get_list(request):
    return CorporationToken.visible_for(request.user)


@router.get("/{int:corporation_id}/", response={200: CorporationTokenSchema, 403: None, 404: None})
def get_single(request, corporation_id: int):
    user: User = request.user

    ownerships = CharacterOwnership.objects.filter(user=user)
    if not user.is_superuser and not ownerships.filter(character__corporation_id=corporation_id).exists():
        return 403, None

    try:
        token = CorporationToken.visible_for(user).get(corporation__corporation_id=corporation_id)
    except CorporationToken.DoesNotExist:
        return 404, None

    return 200, token



from django import forms

from .models import AllianceContact, CorporationContact


class AllianceContactForm(forms.ModelForm):
    class Meta:
        model = AllianceContact
        fields = ('contact_type', 'standing', 'labels', 'notes', )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field, data in self.fields.items():
            if field != 'notes':
                data.disabled = True


class CorporationContactForm(forms.ModelForm):
    class Meta:
        model = CorporationContact
        fields = ('contact_type', 'standing', 'labels', 'notes', )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field, data in self.fields.items():
            if field != 'notes':
                data.disabled = True



# Generated by Django 4.2.14 on 2024-07-23 07:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveonline', '0017_alliance_and_corp_names_are_not_unique'),
        ('aa_contacts', '0003_corporationcontact_is_watched_corporationtoken'),
    ]

    operations = [
        migrations.CreateModel(
            name='StandingFilter',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=500)),
                ('description', models.CharField(max_length=500)),
                ('comparison', models.CharField(choices=[('>', 'Greater Than'), ('>=', 'Greater Or Equal'), ('<', 'Less Than'), ('<=', 'Less Or Equal'), ('=', 'Equal')], max_length=2)),
                ('standing', models.FloatField()),
                ('alliances', models.ManyToManyField(blank=True, help_text='The alliances that have set the standings', related_name='alliance_standing_filters', to='eveonline.eveallianceinfo')),
                ('corporations', models.ManyToManyField(blank=True, help_text='The corporations that have set the standings', related_name='corp_standing_filters', to='eveonline.evecorporationinfo')),
            ],
            options={
                'verbose_name': 'Smart Filter: User Standings',
                'verbose_name_plural': 'Smart Filter: User Standings',
                'default_permissions': (),
            },
        ),
    ]



from django.contrib.auth.models import User

from ninja import Router, Path

from allianceauth.authentication.models import CharacterOwnership

from aa_contacts.models import CorporationToken, CorporationContact
from aa_contacts.tasks import update_corporation_contacts
from ..schema import ContactSchema, UpdateContactSchema

router = Router()


@router.get("/", response={200: list[ContactSchema], 403: None, 404: None})
def list_contacts(request, corporation_id: int = Path(...)):
    user: User = request.user

    ownerships = CharacterOwnership.objects.filter(user=user)
    if not user.is_superuser and not ownerships.filter(character__corporation_id=corporation_id).exists():
        return 403, None

    token = CorporationToken.visible_for(user).filter(corporation__corporation_id=corporation_id)
    if not token.exists():
        return 404, None

    contacts = (
        CorporationContact.objects
        .with_contact_name()
        .filter(corporation__corporation_id=corporation_id)
        .prefetch_related('labels')
    )

    return 200, contacts


@router.post("/update", response={200: None, 403: None, 404: None})
def update_contacts(request, corporation_id: int = Path(...)):
    user: User = request.user

    ownerships = CharacterOwnership.objects.filter(user=user)
    if not user.is_superuser and not ownerships.filter(character__corporation_id=corporation_id).exists():
        return 403, None

    token = (
        CorporationToken.visible_for(user)
        .filter(corporation__corporation_id=corporation_id)
    )
    if not token.exists():
        return 404, None

    update_corporation_contacts(corporation_id)

    return 200, None


@router.patch("/{int:contact_pk}", response={200: None, 403: None, 404: None})
def edit_contact(request, data: UpdateContactSchema, contact_pk: int, corporation_id: int = Path(...)):
    user: User = request.user

    ownerships = CharacterOwnership.objects.filter(user=user)
    if (
        (not user.is_superuser and not ownerships.filter(character__corporation_id=corporation_id).exists())
        or
        (not user.has_perms(['aa_contacts.manage_corporation_contacts', 'aa_contacts.view_corporation_notes']))
    ):
        return 403, None

    token = (
        CorporationToken.visible_for(user)
        .filter(corporation__corporation_id=corporation_id)
    )
    if not token.exists():
        return 404, None

    try:
        contact: CorporationContact = CorporationContact.objects.get(pk=contact_pk, corporation__corporation_id=corporation_id)
    except CorporationContact.DoesNotExist:
        return 404, None

    contact.notes = data.notes
    contact.save(update_fields=['notes'])

    return 200, None



from django.urls import path, re_path

from . import views
from .api import api

app_name = 'aa_contacts'


urlpatterns = [
    path('', views.index, name='index'),
    path('dashboard/', views.dashboard, name='dashboard'),
    path('api/', api.urls),
    re_path('^r/', views.react_view, name='react_view'),
    path('alliance/<int:alliance_pk>/', views.alliance_contacts, name='alliance_contacts'),
    path('alliance/add_token/', views.add_alliance_token, name='add_alliance_token'),
    path('alliance/update/<int:alliance_pk>/', views.update_alliance, name='update_alliance'),
    path('alliance/edit_contact/<int:contact_pk>/', views.update_alliance_contact, name='update_alliance_contact'),
    path('corporation/<int:corporation_pk>/', views.corporation_contacts, name='corporation_contacts'),
    path('corporation/add_token/', views.add_corporation_token, name='add_corporation_token'),
    path('corporation/update/<int:corporation_pk>/', views.update_corporation, name='update_corporation'),
    path('corporation/edit_contact/<int:contact_pk>/', views.update_corporation_contact, name='update_corporation_contact'),
]
